import os
import time
import warnings
import itertools

from collections import defaultdict, Callable
import numpy as np
import sympy
import matplotlib.pyplot as plt
from multiprocessing import Pool
import networkx as nx

from qulacs import Observable, QuantumCircuit, QuantumState
from qulacs.gate import RX, RZ, CNOT, merge


class CircuitByQulacs:
    """generate a instance of CircuitByQulacs"""

    def __init__(self,
                 # p: int = 1,
                 nodes_weight: list = None,
                 edges_weight: list = None,
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByCirq"""

        self._p = None
        self._nodes_weight = nodes_weight
        self._edges_weight = edges_weight
        self._is_parallel = False if is_parallel is None else is_parallel

        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []

    @staticmethod
    def get_operator(element, qubit_num):
        op = Observable(qubit_num)
        if isinstance(element, int):
            op.add_operator(1.0, "Z " + str(element))
        else:
            op.add_operator(1.0, "Z " + str(element[0]) + " Z " + str(element[1]))

        return op

    def get_expectation(self, element_graph, p=None):
        """
        calculate the expectation of the subgraph
        Args:
            element_graph: tuple of (original node/edge, subgraph)

        Returns:
            expectation of the subgraph
        """
        p = self._p if p is None else p
        original_e, graph = element_graph
        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        state = QuantumState(len(node_list))
        state.set_zero_state()

        circ = QuantumCircuit(len(node_list))
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        for k in range(p):
            for i in graph.nodes:
                u = node_to_qubit[i]
                if k == 0:
                    circ.add_H_gate(u)
                circ.add_RZ_gate(u, 2 * gamma_list[k] * self._nodes_weight[i])

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.add_CNOT_gate(u, v)
                circ.add_RZ_gate(v, 2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]])
                circ.add_CNOT_gate(u, v)

            for nd in graph.nodes:
                u = node_to_qubit[nd]
                circ.add_RX_gate(u, 2 * beta_list[k])

        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            op = self.get_operator(node_to_qubit[original_e], len(node_list))
        else:
            weight = self._edges_weight[original_e]
            op = self.get_operator((node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]), len(node_list))

        circ.update_quantum_state(state)
        exp_res = op.get_expectation_value(state)

        return weight * exp_res

    def expectation_calculation(self, p=None):
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        res = 0
        for item in self._element_to_graph.items():
            res += self.get_expectation(item, p)

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        subg_num = len(self._element_to_graph.items())
        args = list(itertools.product(self._element_to_graph.items(), [p] * subg_num))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args), chunksize=1)
        # circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()

        res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="qulacs")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()



# test
# from qulacs import Observable, QuantumCircuit, QuantumState
# from qulacs.gate import Y,CNOT,merge
#
# state = QuantumState(3)
# state.set_Haar_random_state()
#
# circuit = QuantumCircuit(3)
# circuit.add_X_gate(0)
# merged_gate = merge(CNOT(0,1),Y(1))
# circuit.add_gate(merged_gate)
# circuit.add_RX_gate(1,0.5)
# circuit.update_quantum_state(state)
#
# observable = Observable(3)
# observable.add_operator(2.0, "X 2 Y 1 Z 0")
# observable.add_operator(-3.0, "Z 2")
# value = observable.get_expectation_value(state)
# print(value)


""" common module """
import time
import random
from collections import OrderedDict, defaultdict

import numpy as np
import networkx as nx
from qiskit.aqua import aqua_globals
from qiskit.aqua.operators import StateFn


def get_ising_matrix(qubo_mat: np.array):
    """
    calculate the ising matrix according to the Q matrix of problems.

    Args:
        q_mat: the matrix Q in QUBO model of the problem

    Returns:
        ising_mat (np.array): the the matrix of ising model
    """

    ising_mat = np.zeros_like(qubo_mat, dtype='float')  #.copy()
    mat_len = np.size(qubo_mat, 0)
    for i in range(mat_len):
        for j in range(mat_len):
            if i == j:
                ising_mat[i][i] = 1.0 * np.true_divide(qubo_mat[i][i], 2.0) + np.sum(qubo_mat[i]) - qubo_mat[i][i]
            else:
                ising_mat[i][j] = 1.0 * np.true_divide(qubo_mat[i][j], 4.0)

    return ising_mat


def get_weights_graph(ising_mat: np.array, graph: nx.Graph=None):
    """
    use ising matirx as adjacency matrix to generate correspondence graph model
    Args:
        ising_mat (np.array): the Ising matrix that used to generate graph

    Returns:
        graph (nx.Graph): the graph model generated by ising matrix
    """

    cnt = np.size(ising_mat, 0)
    node_map = defaultdict(int)
    if graph is not None:
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_map[i] = node_list[i]
    else:
        for i in range(cnt):
            node_map[i] = i

    graph = nx.Graph()
    for i in range(cnt):
        for j in range(cnt):
            if i == j:
                graph.add_node(node_map[i], weight=ising_mat[i][i])
            else:
                if abs(ising_mat[i][j] - 0.) <= 1e-8:
                    continue
                graph.add_edge(node_map[i], node_map[j], weight=ising_mat[i][j])

    return graph    


def get_most_small_ising(state_count, ising_g):

    nodew = nx.get_node_attributes(ising_g, 'weight')
    edw = nx.get_edge_attributes(ising_g, 'weight')

    min_h = 999999
    for key, val in state_count.items():
        tmp_h = 0
        tw = []
        for i in range(len(key)):
            if key[i] == '0':
                tw.append(-1)
            else:
                tw.append(1)

        for nd in nodew:
            tmp_h += nodew[nd] * tw[nd]
        for ed in edw:
            if ed[0] == ed[1]:
                continue
            tmp_h += edw[ed] * tw[ed[0]] * tw[ed[1]]

        if min_h > tmp_h:
            min_h = tmp_h
            res_state = [int(key[i]) for i in range(len(key))]
    return res_state


def random_regular_graph(node_num, degree=3, weight_range=10, negative_weight=False, seed=None):
    """Generate random Erdos-Renyi graph.

    Args:
        node_num (int): number of nodes.
        weight_range (int): weights will be smaller than this value,
            in absolute value. range: [1, weight_range).
        degree (int): the number of edge belong to every node.
        negative_weight (bool): allow to have edge with negative weights
        # savefile (str or None): name of file where to save graph.
        seed (int or None): random seed - if None, will not initialize.

    Returns:
        numpy.ndarray: adjacency matrix (with weights).

    """
    assert (node_num * degree) % 2 == 0
    assert weight_range >= 0
    if seed is None:
        seed = np.random.randint(1, 10598)   # seed = 10598
        # random.seed(seed)

    random_g = nx.random_regular_graph(d=degree, n=node_num, seed=seed)

    for e in random_g.edges:
        w = random.uniform(1, weight_range)   # np.random.randint(1, weight_range)
        if np.random.random() >= 0.5 and negative_weight:
            w *= -1
        random_g.add_edge(e[0], e[1], weight=w)

    for e in random_g.nodes:
        w = random.uniform(1, weight_range)
        if np.random.random() >= 0.5 and negative_weight:
            w *= -1
        random_g.add_node(e, weight=w)
        random_g.add_edge(e, e, weight=w)
    return random_g


def random_number_list(n, weight_range=(1, 100), seed=None):
    """Generate a set of positive integers within the given range.

    Args:
        n (int): size of the set of numbers.
        weight_range (tuple/list): minimum and maximum absolute value of the numbers.
        savefile (str or None): write numbers to this file.
        seed (Union(int,None)): random seed - if None, will not initialize.

    Returns:
        numpy.ndarray: the list of integer numbers.
    """
    if seed:
        aqua_globals.random_seed = seed

    number_list = aqua_globals.random.integers(low=weight_range[0], high=(weight_range[1] + 1), size=n)
    return number_list


import os
import time
import warnings
from collections import defaultdict, Callable
import numpy as np
import matplotlib.pyplot as plt
from multiprocessing import Pool, cpu_count
import quimb as qu
from multiprocessing import cpu_count

expectation_path = []

class CircuitByTensor:
    """generate a instance of tensor network circuit generated by quimb"""

    def __init__(self,
                 # p: int = 1,
                 nodes_weight: list = None,
                 edges_weight: list = None,
                 contract_opt: str = 'greedy',
                 is_parallel: bool = None) -> None:

        """initialize a instance of tensor network circuit:"""

        # self._p = p
        self._p = None
        self._nodes_weight = nodes_weight
        self._edges_weight = edges_weight
        self._is_parallel = False if is_parallel is None else is_parallel
        self._opt = contract_opt

        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []

    def get_expectation(self, element_graph, p=None):
        p = self._p if p is None else p
        original_e, graph = element_graph
        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ = qu.tensor.Circuit(len(graph.nodes))

        for k in range(p):
            for nd in graph.nodes:
                u = node_to_qubit[nd]
                if k == 0:
                    circ.apply_gate('H',u)
                circ.apply_gate('rz', 2 * gamma_list[k] * self._nodes_weight[nd], u)

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.apply_gate('RZZ', - gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)

            for nd in graph.nodes:
                circ.apply_gate('rx', 2 * beta_list[k], node_to_qubit[nd])

        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            where = node_to_qubit[original_e]
            exp_res = circ.local_expectation(qu.pauli('Z'), where, optimize=self._opt)
        else:
            weight = self._edges_weight[original_e]
            ZZ = qu.pauli('Z') & qu.pauli('Z')
            where = (node_to_qubit[original_e[0]], node_to_qubit[original_e[1]])
            exp_res = circ.local_expectation(ZZ, where, optimize=self._opt)
        return exp_res.real * weight

    def expectation_calculation(self, p=None):
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        cpu_num = cpu_count()
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)
        res = 0
        for item in self._element_to_graph.items():
            res += self.get_expectation(item, p)

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()

        res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="quimb")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()




import numpy as np
from typing import Optional
from scipy import optimize as opt


class Interp:
    """
        Interp optimizer: a heuristic optimization method for QAOA,
        implemented according to the paper
        "Quantum Approximate Optimization Algorithm: Performance, Mechanism, and Implementation on Near-Term Devices"
    """
    def __init__(self,
                 optimize_method='COBYLA',
                 initial_point: Optional[np.ndarray] = None):
        self._p = None
        self._optimize_method = optimize_method
        self._initial_point = initial_point

    def _minimize(self, objective_function):
        """
        minimize the loss function
        Args:
            loss: the loss function
            initial_point: the init parameters of gamma and beta

        Returns:
            x: the optimized gamma and beta
            value: the optimized value of loss function
            nfev: is the number of objective function calls
        """
        nfev = 0
        for k in range(1, self._p + 1):
            if k == 1:
                gamma_list, beta_list = self._initial_point[: k], self._initial_point[self._p : self._p+k]
                gamma_list = np.insert(gamma_list, 0, 0)
                beta_list = np.insert(beta_list, 0, 0)
            else:
                gamma_list = np.insert(gamma_list, 0, 0)
                beta_list = np.insert(beta_list, 0, 0)
                gamma_list = np.append(gamma_list, 0)
                beta_list = np.append(beta_list, 0)
                gamma_list_new, beta_list_new = gamma_list, beta_list
                for i in range(1, k + 1):
                    gamma_list_new[i] = (i - 1) / k * gamma_list[i - 1] + (k - i + 1) / k * gamma_list[i]
                    beta_list_new[i] = (i - 1) / k * beta_list[i - 1] + (k - i + 1) / k * beta_list[i]

                gamma_list, beta_list = gamma_list_new, beta_list_new

            res = opt.minimize(objective_function,
                           x0=np.append(gamma_list[1:k+1], beta_list[1:k+1]),
                           args=k,
                           method=self._optimize_method,
                           tol=1e-6,
                           jac=opt.rosen_der,
                           options={'gtol': 1e-6, 'maxiter': 30, 'disp': True})

            gamma_list, beta_list = res["x"][:k], res["x"][k:]
            value = res["fun"]
            nfev += res["nfev"]
        return res.x, res.fun, res.nfev
        # return {"gamma": gamma_list, "beta": beta_list, "optimal value": value, "nfev": nfev}

    def optimize(self, objective_function, p):

        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * p)])
        return self._minimize(objective_function)


import logging
import numpy as np
import random
import networkx as nx
import matplotlib.pyplot as plt
import time, os
from numbers import Number
import pandas as pd
from numpy.random import choice
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list


logger = logging.getLogger(__name__)


class QadraticKnapsack:
    """
    Qadratic Knapsack problem:
    Given a set of items, each with a weight and a value, 
    determine the number of each item to include in a collection 
    so that the total weight is less than or equal to a given limit 
    and the total value is as large as possible.
    There is an interaction between pairs of projects affecting the value obtained. 
    """
    
    def __init__(self,
                 v: np.array = None,
                 length: int = None,
                 weight_range: tuple = (1, 100),
                 # weight: np.array = None,
                 element_set: list = None, #constriants from subsets
                 b: list = None,
                 P: int = None,
                 slack: int = 3,
                 seed: int = None):
        """
        Args:
            v (np.array): a matrix of values associated with choosing projects 
            length (int): number of variables
            element_set (list) : a list of resource requirement of project
            b (list): a list of the total resource budget 
            P (int): the penalty value for the penalty terms
            slack (int): slack variable, default value is 3
            (require input: v, element_set, b, P, slack(optional))
            
        Returns:
            v, length, element_set, b, P, slack, constraints
            
        Example:
            v_list = [[2,4,3,5],[4,5,1,3],[3,1,2,2],[5,3,2,4]] 
            length = len(v_list)
            subset = [8,6,5,3]
            constant = [16]
            penalty = 10
        """
        self._slack = slack
        
        if v is None and element_set is None: #and distance is None
            assert length is not None
            assert len(element_set) is not None
            
            self._length = length
            self._weight_range = weight_range
            # self._weight = weight
            self._seed = seed
            self._element_set = element_set
            self._b = b
            self._P = P
            self._constraints = self.get_constraints()
            self._v = np.array(random_number_list(n=self._length**2,
                                                   weight_range=self._weight_range,
                                                   seed=self._seed)).reshpae(self._length,self._length)

        else:
            self._v = v
            self._length = len(v)
            self._weight_range = weight_range
            # self._weight = weight
            self._seed = seed
            self._element_set = element_set
            self._b = b
            self._P = P
            self._constraints = self.get_constraints()

        self._qmatrix = None
            
    @property
    def length(self):
        return self._length
    
    def get_constraints(self): 
        """
        convert ineqality constraint into equality constraint which have the form Ax=b
        and get the A matrix from the constraint
        
        Args:
            element_set (list) : a list of resource requirement of project
            slack (int): slack variable coefficient, default value is 3
        Returns: 
            A (np.array): the matrix of constraint coefficients
        
        ..math::
            the slack variable slack default upper bound default value is 3, 
            if slack is an even number, it should be converted to an odd number by adding 1.
            slack_1 is then binary expanded into slack_cof1, 
            A_list is extended by slack_cof1.
        """        
        A_list =[]
        y = list([int(i) for i in bin(self._slack)[2:]]) #slack variable in binary
        y.reverse()
        if self._slack % 2 ==0:
            y = list([int(i) for i in bin(self._slack+1)[2:]])
        slack_cof1 = np.multiply(np.array(y),np.float_power(2, np.arange(len(y)))).tolist()
        A_list = list(self._element_set)
        A_list.extend(int(i) for i in slack_cof1)
        
        return np.array(A_list)
    
    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of set packing problem
        
        Args:
            numbers (np.array): the number set that to be divided
            
        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        
        ..math::
            minimise x(T)Qx
            Q[i][j] = -v[i][j] + P * A[i] * A[j]
            Q[i][i] = -v[i][i] + P * (A[i]**2 - 2 * b * A[i])
            
        """
        matrix_dimension = self._constraints.shape[0]
        q_mat = np.eye(matrix_dimension)
        
        extended_v = [[] for i in range(matrix_dimension)]
        extended_v[0:self._length] = list(self._v)
        for i in range(self._length):
            extended_v[i].extend([0]*(matrix_dimension-self._length))
        for j in range(self._length,matrix_dimension):
            extended_v[j] = [0]*(matrix_dimension)

        for i in range(matrix_dimension):
            for j in range(matrix_dimension):
                if i == j:
                    q_mat[i][i] = -extended_v[i][i] + self._P*((self._constraints[i])**2 - 2*self._b[0]*self._constraints[i]) 
                else:
                    q_mat[i][j] = -extended_v[i][j] + self._P*(self._constraints[i]*self._constraints[j])

        return q_mat

    def quadratic_knapsack_value(self, x,w): 
        """Compute the value of quadratic knapsack.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            weight (numpy.ndarray): weights of elements

        Returns:
            float: value of the knapsack.
        """
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x))) 
        return X
        

    def run(self):
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        qk_graph = get_weights_graph(ising_mat)
        return qk_graph   


import os
import sys
import time
import logging
import random
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_regular_graph


logger = logging.getLogger(__name__)


class SherringtonKirkpatrick:
    """
    Sherrington Kirkpatrick problem: The Sherrington-Kirkpatrick model is a special instance of a spin glass.
    it's defined on the complete graph with wij randomly chosen to be ±1
    """
    def __init__(self, node_num: int = 5):
        self._node_num = node_num

    @property
    def node_num(self):
        return self._node_num

    def run(self):
        ising_mat = np.zeros((self._node_num, self._node_num)) #, dtype=float
        for i in range(self._node_num):
            for j in range(self._node_num):
                if i <= j:
                    ising_mat[i][j] = np.random.choice([1, -1], 1)
                else:
                    ising_mat[i][j] = ising_mat[j][i]
        sk_graph = get_weights_graph(ising_mat)
        return sk_graph



"""
Object to solve QAOA problems

The QAOA problems can be represented as an Ising model, and then be transformed to a DAG.
The directed acyclic graph is decomposed by a specified p value, and these subgraphs then
be transformed as circuits and be executed on simulators, using optimizer to get the
optimal parameters of the original Ising model
"""

import sys
import time
import warnings
from typing import Optional
from collections import defaultdict
import numpy as np
import networkx as nx
from .optimizers import Optimizer, COBYLA
from .backends import Backend, CircuitByQiskit, CircuitByTensor
from .exceptions import GraphTypeError


class Qcover:
    """
    Qcover is a QAOA solver
    """
    # pylint: disable=invalid-name

    def __init__(self,
                 graph: nx.Graph = None,
                 p: int = 1,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQiskit()
                 ) -> None:
        """initialize a instance of Qcover"""

        assert graph is not None
        self._simple_graph = graph
        self._p = p
        self._backend = backend
        self._backend._p = p
        self._optimizer = optimizer
        self._optimizer._p = p

        self._nodes_weight = []
        self._edges_weight = []
        self._hard_to_calcute = False

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, ap):
        self._p = ap

    @property
    def backend(self):
        return self._backend

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def simple_graph(self):
        return self._simple_graph

    @simple_graph.setter
    def simple_graph(self, graph):
        """
        according to the type of graph(nx.graph / tuple) to set the value of
        self._simple_graph
        """
        if isinstance(graph, nx.Graph):
            self._simple_graph = graph
        elif isinstance(graph, tuple):
            assert (len(graph) >= 2) and (len(graph) <= 3)

            if len(graph) == 2:
                node_num, edge_num = graph
                wr = None
            elif len(graph) == 3:
                node_num, edge_num, wr = graph

            nodes, edges = Qcover.generate_graph_data(node_num, edge_num, wr)
            self._simple_graph = self.generate_weighted_graph(nodes, edges)
        elif isinstance(graph, list):
            assert len(graph) == 3
            node_list, edge_list, weight_range = graph
            self._simple_graph = Qcover.generate_weighted_graph(node_list, edge_list, weight_range)
        else:
            print("Error: the argument graph should be a instance of nx.Graph or a tuple formed as (node_num, edge_num)")

    @staticmethod
    def generate_graph_data(node_num, edge_num, weight_range=10):
        """
        generate a simple graph‘s weights of nodes and edges with
        node number is node_num, edge number is edge_num
        Args:
            node_num (int): node number in graph
            edge_num (int): edge number in graph
            weight_range (int): weight range of nodes and edges
        Return:
            nodes(set of tuple(nid, node_weight)), edges(set of tuple(nid1, nid2, edge_weight))
        """
        if weight_range is None:
            weight_range = 10

        nodes = set()
        for i in range(node_num):
            ndw = np.random.choice(range(weight_range))
            nodes |= {(i, ndw)}

        edges = set()
        cnt = edge_num
        max_edges = node_num * (node_num - 1) / 2
        if cnt > max_edges:
            cnt = max_edges
        while cnt > 0:
            u = np.random.randint(node_num)
            v = np.random.randint(node_num)
            if u == v:  # without self loop
                continue
            flg = 0
            for e in edges:  # without duplicated edges
                if set(e[:2]) == set([v, u]):
                    flg = 1
                    break
            if flg == 1:
                continue
            edw = np.random.choice(range(weight_range))
            edges |= {(u, v, edw)}
            cnt -= 1
        return nodes, edges

    @classmethod
    def generate_weighted_graph(cls, nodes, edges, weight_range=10):
        """
        generate graph from nodes list and edges list which identify the nodes and edges
        that should be add in graph, and the random weight range of every node and edge.

        Args:
            nodes (list/set): list of node idex / node-weight map, element form is tuple(nid, weight)
            edges (list/set): list of edge: (e_idex1, e_idex2) / edge-weight map, element form is tuple(nid1, nid2, edge_weight)
            weight_range (int): random weight range of every node and edge

        Returns:
            g (nx.Graph): graph generated by args
        """
        g = nx.Graph()
        if isinstance(nodes, list) and isinstance(edges, list):
            for v in nodes:
                w = np.random.choice(range(weight_range))
                g.add_node(v, weight=w)

            for e in edges:
                w = np.random.choice(range(weight_range))
                g.add_edge(e[0], e[1], weight=w)
        else:
            for item in nodes:
                g.add_node(item[0], weight=item[1])

            for item in edges:
                g.add_edge(item[0], item[1], weight=item[2])
        return g

    def get_graph_weights(self):
        """
        get the weights of nodes and edges in graph
        Args:
            self._simple_graph
        Return:
            node weights form is dict{nid1: node_weight}, edges weights form is dict{(nid1, nid2): edge_weight}
        """
        if self._nodes_weight != [] and self._edges_weight != []:
            return self._nodes_weight, self._edges_weight

        nodew = nx.get_node_attributes(self.simple_graph, 'weight')
        edw = nx.get_edge_attributes(self.simple_graph, 'weight')
        edgew = edw.copy()
        for key, val in edw.items():
            edgew[(key[1], key[0])] = val

        self._nodes_weight = nodew
        self._edges_weight = edgew
        return nodew, edgew

    def generate_subgraph(self, dtype: str, p):
        """
        according to the arguments of dtype and p to generate subgraphs from graph
        Args:
            graph (nx.Graph): graph to be decomposed
            dtype (string): set "node" or "edge", the ways according to which to decompose the graph
            p (int): the p of subgraphs
        Return:
            subg_dict (dict) form as {node_id : subg, ..., (node_id1, node_id2) : subg, ...}
        """
        if dtype not in ["node", "edge"]:
            print("Error: wrong dtype, dtype should be node or edge")
            return None

        if self._nodes_weight == [] or self._edges_weight == []:
            self._nodes_weight, self._edges_weight = self.get_graph_weights()

        subg_dict = defaultdict(list)
        if dtype == 'node':
            for node in self.simple_graph.nodes:
                node_set = {(node, self._nodes_weight[node])}
                edge_set = set()
                for i in range(p):
                    new_nodes = { (nd2, self._nodes_weight[nd2]) for nd1 in node_set for nd2 in self.simple_graph[nd1[0]] }
                    new_edges = {(nd1[0], nd2, self._edges_weight[nd1[0], nd2]) for nd1 in node_set for nd2 in self.simple_graph[nd1[0]]}
                    node_set |= new_nodes
                    edge_set |= new_edges

                subg = self.generate_weighted_graph(node_set, edge_set)
                subg_dict[node] = subg
        else:
            for edge in self.simple_graph.edges:
                node_set = {(edge[0], self._nodes_weight[edge[0]]), (edge[1], self._nodes_weight[edge[1]])}
                edge_set = {(edge[0], edge[1], self._edges_weight[edge[0], edge[1]])}

                for i in range(p):
                    new_nodes = {(nd2, self._nodes_weight[nd2]) for nd1 in node_set for nd2 in self.simple_graph[nd1[0]]}
                    new_edges = {(nd1[0], nd2, self._edges_weight[nd1[0], nd2]) for nd1 in node_set for nd2 in
                                 self.simple_graph.adj[nd1[0]]}
                    node_set |= new_nodes
                    edge_set |= new_edges

                subg = self.generate_weighted_graph(node_set, edge_set)
                subg_dict[edge] = subg
        return subg_dict

    def graph_decomposition(self, p):
        """
        according to dtype to decompose graph
        Args:
            self.simple_graph (nx.Graph): graph to be composed
            self.p (int): the p of subgraphs"""

        if p <= 0:
            warnings.warn(" the argument of p should be >= 1 in qaoa problem, "
                          "so p would be set to the default value at 1")
            p = 1

        subg_node = self.generate_subgraph("node", p)
        subg_edge = self.generate_subgraph("edge", p)
        element_to_graph = {}
        for k, v in subg_node.items():
            element_to_graph[k] = v

        for k, v in subg_edge.items():
            element_to_graph[k] = v
        return element_to_graph

    def calculate(self, pargs, p=None):
        """
        The framework function which use the backend to calculate the value of expectation,
        and be used as the object function in the optimization function of the optimizer
        Args:
            pargs: the value of the parameter alpha and beta in the circuit
            p: the integer used to define the number of layers the current circuit needs to be superimposed

        Returns:
            the value of expectation calculated by backends
        """
        p = self._p if p is None else p
        element_to_graph = self.graph_decomposition(p)

        # checking graph type of given problem
        if not isinstance(self._backend, CircuitByTensor):
            for k, v in element_to_graph.items():
                ncnt, ecnt = len(v.nodes), len(v.edges)
                try:
                    nreq1 = ncnt * (ncnt - 1) <= 2 * ecnt and ncnt >= 20
                    nreq2 = isinstance(k, int) and v.degree[k] >= 30
                    if nreq1 or nreq2:
                        raise GraphTypeError("The problem is transformed into a dense graph, " \
                           "which is difficult to be solved effectively by Qcover")
                except GraphTypeError as e:
                    if not self._hard_to_calcute:
                        print(e)
                        self._hard_to_calcute = True
                    # sys.exit()

        self._backend._pargs = pargs
        self._backend._element_to_graph = element_to_graph
        return self._backend.expectation_calculation(p)

    def run(self, node_num=None, edge_num=None, is_parallel=False):
        """
        run Qcover code to solve the given problem
        Args:
            node_num: nodes number in the graphical representation of the problem
            edge_num: edges number in the graphical representation of the problem
            is_parallel: run programs in parallel or not

        Returns:
            results of the problem, which including the optimal parameters of the circuit model,
            the optimal expectation value and the number of times the optimizer iterates
        """
        if self.simple_graph is None:
            if node_num is None or edge_num is None:
                print("Error: the graph to be solved is not specified, "
                      "arguments of form (node_number, edge_number) should be given")
                return None
            self.simple_graph(node_num, edge_num)

        if self._nodes_weight == [] or self._edges_weight == []:
            self._nodes_weight, self._edges_weight = self.get_graph_weights()

        self._backend._nodes_weight = self._nodes_weight
        self._backend._edges_weight = self._edges_weight
        self._backend._is_parallel = is_parallel

        x, fun, nfev = self._optimizer.optimize(objective_function=self.calculate, p=self._p)
        res = {"Optimal parameter value:": x, "Expectation of Hamiltonian": fun, "Total iterations": nfev}
        return res


# usage example
if __name__ == '__main__':
    node_num, edge_num = 4, 2
    p = 2
    nodes, edges = Qcover.generate_graph_data(node_num, edge_num)
    g = Qcover.generate_weighted_graph(nodes, edges)

    from .optimizers import GradientDescent, Interp, Fourier, COBYLA
    # the numbers in initial_point should be setted by p
    optc = COBYLA(maxiter=30, tol=1e-6, disp=True, initial_point=np.asarray([0.5, 0.5, 0.5, 0.5]))
    optg = GradientDescent(maxiter=50, tol=1e-7, learning_rate=0.0001)
    opti = Interp(optimize_method="COBYLA")  #, initial_point=[0.5, 0.5]
    optf = Fourier(p=p, q=4, r=2, alpha=0.6, optimize_method="COBYLA")

    from .backends import CircuitByQiskit, CircuitByCirq, CircuitByQulacs, CircuitByProjectq, CircuitByTensor
    qiskit_bc = CircuitByQiskit(expectation_calc_method="statevector")   # sample
    ts_bc = CircuitByTensor()
    cirq_bc = CircuitByCirq()
    qulacs_bc = CircuitByQulacs()
    pq_bc = CircuitByProjectq()

    qser_sta = Qcover(g, p,
                      optimizer=optf,
                      backend=ts_bc)  #qulacs_bc cirq_bc pq_bc qiskit_bc

    time_start = time.time()
    res_sta_np = qser_sta.run(is_parallel=False)  # True
    time_end = time.time()
    print('statevector without parallel takes: ', time_end - time_start)
    print("the result from optimizer is:\n", res_sta_np)
    qser_sta.backend.visualization()



import os
import sys
import logging
import time
import numpy as np
import networkx as nx
import random
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_regular_graph


logger = logging.getLogger(__name__)


class MaxCut:
    """
    max cut problem:
    Divide the vertex set V of the graph into two sets so that the number of edges connecting
    the two vertex sets is as large as possible.
    For the weight graph, the weight sum of the edges connecting the two sets is required to be the largest.
    """
    def __init__(self,
                 graph: nx.Graph = None,
                 node_num: int = None,
                 node_degree: int = 3,
                 weight_range: int = 10,
                 seed: int = None) -> None:

        if graph is None:
            # generate random graph according to node_num and node_degree and weight_range
            self._node_num = node_num
            self._graph = random_regular_graph(node_num=self._node_num,
                                               degree=node_degree,
                                               weight_range=weight_range,
                                               seed=seed)
        else:
            self._node_num = len(graph.nodes)
            self._graph = graph

        self._qmatrix = None

    @property
    def node_num(self):
        return self._node_num

    @property
    def graph(self):
        return self._graph

    def update_random_graph(self, node_num, node_degree, weight_range, seed):
        self._node_num = node_num
        self._graph = random_regular_graph(node_num=self._node_num,
                                           degree=node_degree,
                                           weight_range=weight_range,
                                           seed=seed)

    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of max cut problem
        Args:
            adjacent_mat (np.array): the adjacent matrix of graph G of the problem

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        """
        adj_mat = nx.adjacency_matrix(self._graph).A
        qubo_mat = adj_mat.copy()
        for i in range(self._node_num):
            qubo_mat[i][i] = - (np.sum(adj_mat[i]) - adj_mat[i][i])

        return qubo_mat

    def max_cut_value(self, x, w):
        """Compute the value of a cut.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            w (numpy.ndarray): adjacency matrix.

        Returns:
            float: value of the cut.
        """
        X = np.outer(x, (1 - x))
        return np.sum(w * X)

    def run(self):
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        mc_graph = get_weights_graph(ising_mat)
        return mc_graph




import os
import sys
import logging
import time
import numpy as np
import networkx as nx
import random
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list


logger = logging.getLogger(__name__)


class NumberPartition:
    """
    number partition problem:
    given the set of numbers, divided it into two parties
    so the sum of numbers in two parties are as close as possible
    """
    def __init__(self,
                 number_list: np.array = None,
                 length: int = None,
                 weight_range: tuple = (1, 100),
                 seed: int = None):

        if number_list is None:
            assert length is not None

            self._length = length
            self._weight_range = weight_range
            self._seed = seed
            self._number_list = random_number_list(n=self._length,
                                                   weight_range=self._weight_range,
                                                   seed=self._seed)
        else:
            self._number_list = number_list
            self._length = len(number_list)
            self._weight_range = (np.abs(number_list).min(), np.abs(number_list).max())
            self._seed = seed

    @property
    def length(self):
        return self._length

    @property
    def weight_range(self):
        return self._weight_range

    @property
    def number_list(self):
        return self._number_list

    def update_args(self, length, weight_range):
        self._length = length
        self._weight_range = weight_range

        self._number_list = random_number_list(n=self._length, weight_range=self._weight_range, seed=self._seed)   # self.get_random_number_list()

    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of number partition problem

        Args:
            numbers (np.array): the number set that to be divided

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        """

        all_sum = np.sum(self._number_list)
        q_mat = np.eye(self._length)

        for i in range(self._length):
            for j in range(self._length):
                if i == j:
                    q_mat[i][i] = self._number_list[i] * (self._number_list[i] - all_sum)
                else:
                    q_mat[i][j] = self._number_list[i] * self._number_list[j]

        return q_mat

    def partition_value(self, x, number_list):  #
        """Compute the value of a partition.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            number_list (numpy.ndarray): list of numbers in the instance.

        Returns:
            float: difference squared between the two sides of the number
                partition.
        """
        diff = np.sum(number_list[x == 0]) - np.sum(number_list[x == 1])
        return diff * diff

    def run(self):
        qubo_mat = self.get_Qmatrix()
        ising_mat = get_ising_matrix(qubo_mat)
        np_graph = get_weights_graph(ising_mat, )
        return np_graph





# This code is part of Qiskit.
#
# (C) Copyright IBM 2018, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# Notice: This document has been modified on the basis of the original code

import numpy as np
from typing import Iterator, Optional, Union, Callable, Dict, Any
from functools import partial

CALLBACK = Callable[[int, np.ndarray, float, float], None]


class GradientDescent:

    def __init__(self,
            # p: int = 1,
            maxiter: int = 100,
            learning_rate: Union[float, Callable[[], Iterator]] = 0.0005,
            tol: float = 1e-7,
            callback: Optional[CALLBACK] = None,
            perturbation: Optional[float] = None,
            initial_point: Optional[np.ndarray] = None) -> None:

        self._p = None
        self.maxiter = maxiter
        self.learning_rate = learning_rate
        self.perturbation = perturbation
        self.tol = tol
        self.callback = callback
        self._max_evals_grouped = 1
        self._initial_point = initial_point

    @property
    def settings(self) -> Dict[str, Any]:
        # if learning rate or perturbation are custom iterators expand them
        if callable(self.learning_rate):
            iterator = self.learning_rate()
            learning_rate = np.array([next(iterator) for _ in range(self.maxiter)])
        else:
            learning_rate = self.learning_rate

        return {
            "maxiter": self.maxiter,
            "tol": self.tol,
            "learning_rate": learning_rate,
            "perturbation": self.perturbation,
            "callback": self.callback,
        }

    @staticmethod
    def gradient_num_diff(x_center, f, epsilon, max_evals_grouped=1):
        """
        We compute the gradient with the numeric differentiation in the parallel way,
        around the point x_center.

        Args:
            x_center (ndarray): point around which we compute the gradient
            f (func): the function of which the gradient is to be computed.
            epsilon (float): the epsilon used in the numeric differentiation.
            max_evals_grouped (int): max evals grouped
        Returns:
            grad: the gradient computed

        """
        forig = f(*((x_center,)))
        grad = []
        ei = np.zeros((len(x_center),), float)
        todos = []
        for k in range(len(x_center)):
            ei[k] = 1.0
            d = epsilon * ei
            todos.append(x_center + d)
            ei[k] = 0.0

        counter = 0
        chunk = []
        chunks = []
        length = len(todos)
        # split all points to chunks, where each chunk has batch_size points
        for i in range(length):
            x = todos[i]
            chunk.append(x)
            counter += 1
            # the last one does not have to reach batch_size
            if counter == max_evals_grouped or i == length - 1:
                chunks.append(chunk)
                chunk = []
                counter = 0

        for chunk in chunks:  # eval the chunks in order
            parallel_parameters = np.concatenate(chunk)
            todos_results = f(parallel_parameters)  # eval the points in a chunk (order preserved)
            if isinstance(todos_results, float):
                grad.append((todos_results - forig) / epsilon)
            else:
                for todor in todos_results:
                    grad.append((todor - forig) / epsilon)

        return np.array(grad)

    def _minimize(self, loss, grad):   #, initial_point
        # set learning rate
        if isinstance(self.learning_rate, float):
            eta = constant(self.learning_rate)
        else:
            eta = self.learning_rate()

        if grad is None:
            eps = 0.01 if self.perturbation is None else self.perturbation
            grad = partial(
                GradientDescent.gradient_num_diff,
                f=loss,
                epsilon=eps,
                max_evals_grouped=self._max_evals_grouped,
            )

        # prepare some initials
        x = np.asarray(self._initial_point)
        nfevs = 0

        for _ in range(1, self.maxiter + 1):
            # compute update -- gradient evaluation counts as one function evaluation
            update = grad(x)
            nfevs += 1

            # compute next parameter value
            x_next = x - next(eta) * update

            # send information to callback
            stepsize = np.linalg.norm(update)
            if self.callback is not None:
                self.callback(nfevs, x_next, loss(x_next), stepsize)

            # update parameters
            x = x_next

            # check termination
            if stepsize < self.tol:
                break

        return x, loss(x), nfevs

    def optimize(self, objective_function, p, gradient_function=None):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * p)])
        return self._minimize(objective_function, gradient_function)  #, initial_point


def constant(eta=0.01):
    """Yield a constant."""

    while True:
        yield eta



import logging
import numpy as np
import random
import networkx as nx
import matplotlib.pyplot as plt
import time, os
from numbers import Number
import pandas as pd
from numpy.random import choice
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list


logger = logging.getLogger(__name__)


class SetPartitioning:
    """
    Set Partitioning problem:
    For a finite set S, partitioning S into several subsets such that each element of S is in one and only one subset,
    find the partition that minimises the cost without violating any constraint.
    
    """
    def __init__(self,
                 element_list: list = None,
                 length: int = None,
                 weight_range: tuple = (1, 100),
                 element_set: np.array = None, #constriants from subsets
                 weight: np.array = None,
                 P: int = None,
                 seed: int = None):
        """
        Args:
            element_list (list): a list of elements 
            length (int): length of number_list
            element_set (list): a list of constraints
            weight (list): list of weight for element in number_list
            P (int): the penalty value for the penalty terms
            (require input: element_list, element_set, weight, P)
            
        Returns: 
           element_list, length, weight_range, weight, element_set, constraints, P
            
        Example:
            element_list = ['a','b','c','d','e','f']
            element_list_len = len(element_list)
            element_weight = [3,2,1,1,3,2]
            subsets = [[1,3,6],[2,3,5,6],[3,4,5],[1,2,4,6]]
            penalty = 10
        """
                
        if element_list is None and element_set is None:
            assert length is not None
            assert len(element_set) is not None
            
            self._length = length
            self._weight_range = weight_range
            self._weight = weight
            self._element_set = element_set
            self._constraints = self.constraints()
            self._P = P
            self._seed = seed
            self._element_list = random_number_list(n=self._length,
                                                   weight_range=self._weight_range,
                                                   seed=self._seed)
            
        else:
            
            for n in [element_list]:
                if isinstance(n, Number) == False:
                    self._element_list = pd.factorize(element_list)[0] 
            self._length = len(self._element_list)
            self._weight_range = (np.abs(self._element_list).min(), np.abs(self._element_list).max())
            self._seed = seed
            self._weight = weight
            self._element_set = element_set
            self._constraints = self.constraints()
            self._P = P

        self._qmatrix = None
            
    @property
    def length(self):
        return self._length

    @property
    def weight(self):
        return self._weight
    
    def constraints(self): 
        """
        If the element in the list is included in the subset, 
        the constraints matrix entry is 1, vice versa
        """
        constraints_mat = np.zeros((self._length,self._length), dtype=int)
        for i in range(self._length): #list length
            for j in range(self._length):
                for a in range(len(self._element_set)): #number of constraints
                    if i+1 in self._element_set[a] and j+1 in self._element_set[a]: 
                        constraints_mat[a][i] = 1
        return constraints_mat #binary matrix

    def update_args(self, length, weight,constraints):
        self._length = length
        self._weight = weight
        self._constraints = constraints
    
    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of number partition problem

        Args:
            numbers (np.array): the number set that to be divided

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        
        ..math::
        minimise x(T)Qx
        Q[i][j] = P * number of times elements i and j appear in the same constraint
        Q[i][i] = -weight[i] - P * number of ones in the ith column of the constraint matrix
        """
        
        q_mat = np.eye(self._length)
        
        for i in range(self._length):
                q_mat[i][i] = self._weight[i] - self._P * np.sum(self._constraints, axis = 0)[i]                           
                for j in range(i):
                    r = 0 
                    for a in range(len(self._element_set)): #number of constraints
                        if i+1 in self._element_set[a] and j+1 in self._element_set[a]: 
                            r += 1 #number of times elements i and j appear in the same constraint
                            q_mat[i][j] = self._P  * r
                            q_mat[j][i] = q_mat[i][j]

        return q_mat

    def set_partitioning_value(self, x,w): 
        """Compute the value of a packing.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            weight (numpy.ndarray): weights of elements

        Returns:
            float: value of the partitioning.
        """
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()
            
        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x)) ) 
        return X

    def run(self):
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        sp_graph = get_weights_graph(ising_mat)
        return sp_graph


"""Exceptions for errors raised by Qcover."""

from typing import Optional
import warnings


class QcoverError(Exception):
    """Base class for errors raised by Qcover."""

    def __init__(self, *message):
        """Set the error message."""
        super().__init__(" ".join(message))
        self.message = " ".join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)


class GraphTypeError(QcoverError):
    """Raised when a sequence subscript is out of range."""

    def __init__(self, msg):   #, *args
        """Set the error message."""
        # warnings.warn(
        # "The problem is transformed into a dense graph, which is difficult to be solved effectively by Qcover",
        #     stacklevel=2,
        # )
        # self.message = "The problem is transformed into a dense graph, " \
        #                "which is difficult to be solved effectively by Qcover"
        self.name = "GraphTypeError: "
        self.message = msg
        super().__init__(msg)

    def __str__(self):
        return self.name + self.message
        # print("The problem is transformed into a dense graph, which is difficult to be solved effectively by Qcover")


class QQcoverUserConfigError(QcoverError):
    """Raised when an error is encountered reading a user config file."""

    message = "User config invalid"


import platform
from codecs import open
from os import path

#from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# def _process_requirements():
#     packages = open('requirements.txt', "r",  encoding="utf-16").read().strip().split('\n')
#     requires = []
#     for pkg in packages:
#         if pkg.startswith('git+ssh'):
#             return_code = os.system('pip install {}'.format(pkg))
#             assert return_code == 0, 'error, status_code is: {}, exit!'.format(return_code)
#         else:
#             requires.append("\"" + str(pkg) + "\"")
#     return requires

# This reads the __version__ variable from Qcover/version.py
__version__ = ""
exec(open('Qcover/version.py').read())

requirements = [
    "networkx==2.5.1",
    "qiskit==0.31.0",
    "projectq==0.6.1.post0",
    "cirq==0.13.0",
    "quimb==1.3.0",
    "qulacs==0.3.0",
]

setup(
    name="Qcover",
    version=__version__,
    author="ntyz&finleyzhuang",
    author_email="puyn@baqis.ac.cn",
    description="Quantum computing solver",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="Apache-2.0 License",
    url="https://github.com/BAQIS-Quantum/Qcover",
    keywords="QAOA based combinational optimization solver QUBO quantum",

    packages=find_packages(),
    #ext_modules = cythonize("waveforms/math/prime.pyx"),
    include_package_data=True,
    #data_files=[('waveforms/Data', waveData)],
    install_requires=requirements,
    extras_require={
        'test': [
            'pytest>=4.4.0',
        ],
        'docs': [
            'Sphinx',
            'sphinxcontrib-napoleon',
            'sphinxcontrib-zopeext',
        ],
    },
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/BAQIS-Quantum/Qcover/issues',
        'Source': 'https://github.com/BAQIS-Quantum/Qcover',
    },
)


from typing import Optional
import logging
import numpy as np
from scipy import optimize as opt

logger = logging.getLogger(__name__)


class COBYLA:
    """
    COBYLA: a numerical optimization method for constrained problems

    based on scipy.optimize.minimize COBYLA.
    For further detail, please refer to
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    """

    # pylint: disable=unused-argument
    def __init__(self,
                 maxiter: int = 30,
                 initial_point: Optional[np.ndarray] = None,
                 disp: bool = False,
                 rhobeg: float = 1.0,
                 tol: Optional[float] = 1e-6) -> None:
        """
        Args:
            maxiter: Maximum number of function evaluations.
            disp: Set to True to print convergence messages.
            rhobeg: Reasonable initial changes to the variables.
            tol: Final accuracy in the optimization (not precisely guaranteed).
                 This is a lower bound on the size of the trust region.
        """
        self._p = None
        self._tol = tol
        self._options = {'rhobeg': rhobeg, 'maxiter': maxiter, 'disp': disp}
        self._initial_point = initial_point

    def optimize(self, objective_function, p):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * p)])

        res = opt.minimize(objective_function,
                           x0=np.array(self._initial_point),
                           args=p,
                           method='COBYLA',
                           tol=self._tol,
                           jac=opt.rosen_der,
                           options=self._options)
        return res.x, res.fun, res.nfev


import os
import time
import warnings
from collections import defaultdict, Callable
import numpy as np
import matplotlib.pyplot as plt
from multiprocessing import Pool, cpu_count

from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, Aer, assemble, BasicAer, transpile
from qiskit.aqua import QuantumInstance, aqua_globals
from qiskit.aqua.operators import PauliExpectation, CircuitSampler, StateFn, CircuitOp, CircuitStateFn, \
    MatrixExpectation, X, Y, Z, I

expectation_path = []

class CircuitByQiskit:
    """generate a instance of CircuitByQiskit"""
    def __init__(self,
                 # p: int = 1,
                 nodes_weight: list = None,
                 edges_weight: list = None,
                 expectation_calc_method: str = "statevector",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByQiskit"""

        self._p = None
        self._nodes_weight = nodes_weight
        self._edges_weight = edges_weight
        self._is_parallel = False if is_parallel is None else is_parallel
        self._expectation_calc_method = expectation_calc_method

        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        # self._subg_to_circuit = None

    def get_operator(self, element, qubit_num):
        if self._expectation_calc_method == "statevector":
            op = 1
            for i in range(qubit_num):
                if i in list(element):
                    op = np.kron(Z.to_matrix(), op)
                else:
                    op = np.kron(I.to_matrix(), op)
        else:
            op = Z if 0 in list(element) else I
            for i in range(1, qubit_num):
                if i in list(element):
                    op = Z ^ op
                else:
                    op = I ^ op
        return op

    def get_expectation(self, element_graph, p=None):
        p = self._p if p is None else p
        original_e, graph = element_graph
        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ_qreg = QuantumRegister(len(graph), 'q')
        circ = QuantumCircuit(circ_qreg)

        for k in range(p):
            for nd in graph.nodes:
                u = node_to_qubit[nd]
                if k == 0:
                    circ.h(u)
                circ.rz(2 * gamma_list[k] * self._nodes_weight[nd], u)

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)

            for nd in graph.nodes:
                circ.rx(2 * beta_list[k], node_to_qubit[nd])

        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            op = self.get_operator([node_to_qubit[original_e]], len(node_list))
        else:
            weight = self._edges_weight[original_e]
            op = self.get_operator([node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]], len(node_list))

        if self._expectation_calc_method == "statevector":
            circ.save_statevector()
            sim = Aer.get_backend('aer_simulator')
            result = sim.run(circ).result()
            out_state = result.get_statevector()
            exp_res = np.matmul(np.matmul(out_state.conj().T, op), out_state).real
        else:
            subc = CircuitStateFn(circ)
            backend = Aer.get_backend('qasm_simulator')
            q_instance = QuantumInstance(backend, shots=1024)
            measurable_expression = StateFn(op, is_measurement=True).compose(subc)
            expectation = PauliExpectation().convert(measurable_expression)
            sampler = CircuitSampler(q_instance).convert(expectation)
            exp_res = sampler.eval().real

        return weight * exp_res

    def expectation_calculation(self, p=None):
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        cpu_num = cpu_count()
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)
        res = 0
        for item in self._element_to_graph.items():
            res += self.get_expectation(item, p)

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()

        res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="Qiskit")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()

    # below is functions not used now
    def graph_to_circuit(self, params, graph, p=None, original_e=None):
        """
        transform the graph to circuit according to the computing_framework
        Args:
            graph (nx.Graph): graph to be transformed to circuit
            params (List): Optimal parameters   #np.array
            original_e (Optional[None, int, tuple])
        Return:
            if original_e=None, then the graph is the whole original graph generated by
            generate_weighted_graph(), so just return the circuit transformed by it

            if original_e is a int, then the subgraph is generated by node(idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[], and the circuit

            if original_e is a tuple, then the subgraph is generated by edge(node idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[] as
            tuple(mapped node_id1, mapped node_id2), and the circuit
        """

        # generate node_to_qubit[] map from node id in graph to [0, node_num in graph - 1]
        # so the subgraph would has it's own qubits number
        p = self._p if p is None else p

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        circ_qreg = QuantumRegister(len(graph), 'q')
        circ = QuantumCircuit(circ_qreg)

        gamma_list, beta_list = params[:p], params[p:]
        # generate circuit from graph
        # circuit should repeat self._p times with different gamma & beta every time
        for k in range(p):
            for i in graph.nodes:
                u = node_to_qubit[i]
                if k == 0:
                    circ.h(u)
                circ.rz(2 * gamma_list[k] * self._nodes_weight[i], u)

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)

            for i in graph.nodes:
                circ.rx(2 * beta_list[k], node_to_qubit[i])
        if original_e is None:
            return original_e, circ
        elif isinstance(original_e, int):
            return node_to_qubit[original_e], circ
        else:
            return (node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]), circ

    def get_result_statevector(self, params, graph):

        params = params.flatten()
        _, circ = self.graph_to_circuit(params, graph)
        circ.save_statevector()
        # subc.draw(output='mpl', interactive=True)

        sim = Aer.get_backend('aer_simulator')
        result = sim.run(circ).result()
        out_state = result.get_statevector()

        return out_state

    def get_result_counts(self, params, graph):  # 直接将最优参数值带入整个大电路中运行qubit将只能到18

        params = params.flatten()
        _, circ = self.graph_to_circuit(params, graph)
        circ.measure_all()   # if use sample method, can't use measure

        sim = Aer.get_backend('qasm_simulator')
        transpiled_qc = transpile(circ, sim)
        job_sim = sim.run(transpiled_qc, shots=1024)
        hist = job_sim.result().get_counts(transpiled_qc)
        # for key, val in hist.items():
        #     print("%s has %s" % (key, str(val)))
        # plot_histogram(hist)
        # plt.show()
        return hist




from .optimizer import Optimizer
from .cobyla import COBYLA
from .gradient_descent import GradientDescent
from .interp import Interp
from .fourier import Fourier

__all__ = ['Optimizer',
           'COBYLA',
           'GradientDescent',
           'Interp',
           'Fourier'
           ]